#
# Copyright 2019-2021 Canonical Ltd.
# Authors:
# - dann frazier <dann.frazier@canonical.com>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#

import enum
import os
from pathlib import Path
import shutil
import tempfile
from util import delete_tempfile

class QemuEfiMachine(enum.Enum):
    OVMF_PC = enum.auto()
    OVMF_Q35 = enum.auto()
    AAVMF = enum.auto()
    RISCV64 = enum.auto()
    LOONGARCH64 = enum.auto()


class QemuEfiVariant(enum.Enum):
    MS = enum.auto()
    SECBOOT = enum.auto()
    MS_STRICTNX = enum.auto()
    SNAKEOIL = enum.auto()
    LEGACY = enum.auto()


class QemuEfiFlashSize(enum.Enum):
    DEFAULT = enum.auto()
    SIZE_4MB = enum.auto()


class QemuCommand:
    Qemu_Common_Params = [
        '-no-user-config', '-nodefaults',
        '-m', '256',
        '-smp', '1,sockets=1,cores=1,threads=1',
        '-display', 'none',
        '-serial', 'stdio',
    ]
    Aavmf_Common_Params = Qemu_Common_Params + [
        '-machine', 'virt', '-device', 'virtio-serial-device',
    ]
    LoongArch_Common_Params = Qemu_Common_Params + [
        '-machine', 'virt',
    ]
    Ovmf_Common_Params = Qemu_Common_Params + [
        '-chardev', 'pty,id=charserial1',
        '-device', 'isa-serial,chardev=charserial1,id=serial1',
    ]
    RiscV_Common_Params = Qemu_Common_Params + [
        '-machine', 'virt', '-device', 'virtio-serial-device',
    ]
    Machine_Base_Command = {
        QemuEfiMachine.AAVMF: [
            'qemu-system-aarch64', '-cpu', 'cortex-a57',
        ] + Aavmf_Common_Params,
        QemuEfiMachine.LOONGARCH64: [
            'qemu-system-loongarch64',
        ] + LoongArch_Common_Params,
        QemuEfiMachine.OVMF_PC: [
            'qemu-system-x86_64', '-machine', 'pc,accel=tcg',
        ] + Ovmf_Common_Params,
        QemuEfiMachine.OVMF_Q35: [
            'qemu-system-x86_64', '-machine', 'q35,accel=tcg',
        ] + Ovmf_Common_Params,
        QemuEfiMachine.RISCV64: [
            'qemu-system-riscv64',
        ] + RiscV_Common_Params,
    }

    def _get_default_flash_paths(self, machine, variant, flash_size):
        assert(machine in QemuEfiMachine)
        assert(variant is None or variant in QemuEfiVariant)
        assert(flash_size in QemuEfiFlashSize)

        root = Path(os.environ.get('DEB_EDK2_ROOT', '/'))

        code_ext = '.no-secboot' if machine == QemuEfiMachine.AAVMF else ''
        vars_ext = ''
        if variant == QemuEfiVariant.MS:
            code_ext = vars_ext = '.ms'
        elif variant == QemuEfiVariant.SECBOOT:
            code_ext = '.secboot'
        elif variant == QemuEfiVariant.MS_STRICTNX:
            code_ext = '.secboot.strictnx'
            vars_ext = '.ms'
        elif variant == QemuEfiVariant.SNAKEOIL:
            code_ext = vars_ext = '.snakeoil'

        if variant == QemuEfiVariant.SNAKEOIL:
            # We provide one size - you don't get to pick.
            assert(flash_size == QemuEfiFlashSize.DEFAULT)

        if variant == QemuEfiVariant.LEGACY:
            return (root / 'usr/share/ovmf/OVMF.legacy.fd', None)

        if machine == QemuEfiMachine.AAVMF:
            assert(flash_size == QemuEfiFlashSize.DEFAULT)
            return (
                root / f'usr/share/AAVMF/AAVMF_CODE{code_ext}.fd',
                root / f'usr/share/AAVMF/AAVMF_VARS{vars_ext}.fd'
            )
        if machine == QemuEfiMachine.LOONGARCH64:
            assert(variant is None)
            assert(flash_size == QemuEfiFlashSize.DEFAULT)
            return (
                root / 'usr/share/qemu-efi-loongarch64/QEMU_EFI.fd',
                root / 'usr/share/qemu-efi-loongarch64/QEMU_VARS.fd'
            )
        if machine == QemuEfiMachine.RISCV64:
            assert(variant is None)
            assert(flash_size == QemuEfiFlashSize.DEFAULT)
            return(
                root / 'usr/share/qemu-efi-riscv64/RISCV_VIRT_CODE.fd',
                root / 'usr/share/qemu-efi-riscv64/RISCV_VIRT_VARS.fd',
            )
        # Remaining possibilities are OVMF variants
        assert(
            flash_size in [
                QemuEfiFlashSize.DEFAULT, QemuEfiFlashSize.SIZE_4MB
            ]
        )
        size_ext = '_4M'
        if machine in [QemuEfiMachine.OVMF_PC]:
            assert(variant is None)
        return (
            root / f'usr/share/OVMF/OVMF_CODE{size_ext}{code_ext}.fd',
            root / f'usr/share/OVMF/OVMF_VARS{size_ext}{vars_ext}.fd',
        )

    def __init__(
            self, machine, variant=None,
            code_path=None, vars_template_path=None,
            flash_size=QemuEfiFlashSize.DEFAULT,
    ):
        assert(
            (code_path and vars_template_path) or
            (not code_path and not vars_template_path)
        )

        if not code_path:
            (code_path, vars_template_path) = self._get_default_flash_paths(
                machine, variant, flash_size)

        self.pflash = self.PflashParams(code_path, vars_template_path, variant == QemuEfiVariant.LEGACY)
        self.command = self.Machine_Base_Command[machine] + self.pflash.params

    def add_disk(self, path, pvscsi=False):
        if pvscsi:
            self.command += [
                '-device', 'pvscsi,id=scsi',
                '-drive', f'file={path},if=none,format=raw,id=drive0',
                '-device', 'scsi-hd,drive=drive0,bus=scsi.0'
            ]
        else:
            self.command += [
                '-drive', f'file={path},format=raw'
            ]

    def add_oem_string(self, type, string):
        string = string.replace(",", ",,")
        self.command = self.command + [
            '-smbios', f'type={type},value={string}'
        ]

    class PflashParams:
        '''
        Used to generate the appropriate -pflash arguments for QEMU. Mostly
        used as a fancy way to generate a per-instance vars file and have it
        be automatically cleaned up when the object is destroyed.
        '''
        def __init__(self, code_path, vars_template_path, unified=False):
            if unified:
                self.unified = tempfile.NamedTemporaryFile(
                    prefix=os.path.basename(code_path),
                    delete=delete_tempfile(),
                )
                with open(code_path, 'rb') as template:
                    shutil.copyfileobj(template, self.unified)
                    self.unified.flush()
            self.params = [
                '-drive',
                'file=%s,if=pflash,format=raw,unit=0,readonly=on' %
                (self.unified.name if unified else code_path),
            ]
            if vars_template_path is None:
                return
            self.varfile = tempfile.NamedTemporaryFile(
                prefix=os.path.basename(vars_template_path),
                delete=delete_tempfile(),
            )
            with open(vars_template_path, 'rb') as template:
                shutil.copyfileobj(template, self.varfile)
                self.varfile.flush()
                self.params = self.params + [
                    '-drive',
                    'file=%s,if=pflash,format=raw,unit=1,readonly=off' %
                    (self.varfile.name)
                ]



#
# Copyright 2022 Canonical Ltd.
# Authors:
# - dann frazier <dann.frazier@canonical.com>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import tempfile
from util import dbg_check_call
from util import delete_tempfile

class SignedBinary:
    def __init__(self, binary_path, key_path, cert_path, password=None):
        self.signed_bin = tempfile.NamedTemporaryFile(
                prefix=os.path.basename(binary_path),
                delete=delete_tempfile(),
        )
        self.path = self.signed_bin.name

        openssl_password_args = []
        if password:
            openssl_password_args = [
                "-passin", f"pass:{password}"
            ]
        with tempfile.NamedTemporaryFile(
                prefix=__class__.__name__,
                delete=delete_tempfile(),
        ) as keytmp:
            dbg_check_call(
                [
                    "openssl", "rsa",
                ] + openssl_password_args + [
                    "-in", f"{key_path}",
                    "-out", f"{keytmp.name}",
                ]
            )
            dbg_check_call(
                [
                    "sbsign", "--key", f"{keytmp.name}",
                    "--cert", f"{cert_path}",
                    binary_path, "--output", f"{self.signed_bin.name}"
                ]
            )



#!/usr/bin/env python3
#
# Copyright 2026 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 or later, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Integration tests for the FirmwareSecvarUpdater DXE driver.

Tests that booting firmware with only 2011 Microsoft CAs enrolled
results in the 2023 CAs being appended to KEK and db.
"""

import logging
import os
import subprocess
import sys
import tempfile
import time
import unittest
from collections import UserDict
from pathlib import Path

import pexpect
from virt.firmware.varstore import autodetect

from UEFI.Qemu import QemuEfiMachine, QemuEfiFlashSize
from UEFI import Qemu

TEST_TIMEOUT = 120

# CN strings found in the DER encoding of each certificate.
CERT_CNS = {
    'kek_2011':    b'Microsoft Corporation KEK CA 2011',
    'kek_2023':    b'Microsoft Corporation KEK 2K CA 2023',
    'pca_2011':    b'Microsoft Windows Production PCA 2011',
    'uefi_2011':   b'Microsoft Corporation UEFI CA 2011',
    'uefi_2023':   b'Microsoft UEFI CA 2023',
    'oprom_2023':  b'Microsoft Option ROM UEFI CA 2023',
    'win_2023':    b'Windows UEFI CA 2023',
}

ALL_2023_CNS = [
    CERT_CNS['kek_2023'],
    CERT_CNS['uefi_2023'],
    CERT_CNS['oprom_2023'],
    CERT_CNS['win_2023'],
]

# Expected state after a successful update: every cert that should
# be present in KEK and db, keyed by (var_name, cert_key).
EXPECTED_CERTS = [
    ('KEK', 'kek_2011'),
    ('KEK', 'kek_2023'),
    ('db',  'pca_2011'),
    ('db',  'uefi_2011'),
    ('db',  'win_2023'),
    ('db',  'uefi_2023'),
    ('db',  'oprom_2023'),
]


def has_cert_cn(varstore_path, var_name, cn_bytes):
    """Check if the variable's raw data contains the given CN string."""
    vs = autodetect.open_varstore(varstore_path)
    varlist = vs.get_varlist()
    if var_name not in varlist:
        return False
    return cn_bytes in varlist[var_name].data


def strip_2023_certs(varstore_path, output_path):
    """
    Copy a varstore, removing 2023 Microsoft certificates from KEK
    and db to simulate a legacy VM with only 2011 CAs.
    Returns the number of signature list entries removed.
    """
    vs = autodetect.open_varstore(varstore_path)
    varlist = vs.get_varlist()

    removed = 0
    for var_name in ('KEK', 'db'):
        if var_name not in varlist:
            continue
        kept = []
        for siglist in varlist[var_name].sigdb:
            cert_data = siglist[0]['data'] if len(siglist) > 0 else b''
            if any(cn in cert_data for cn in ALL_2023_CNS):
                removed += 1
            else:
                kept.append(siglist)
        new_sigdb = varlist[var_name].sigdb.__class__()
        new_sigdb.extend(kept)
        varlist[var_name].sigdb_set(new_sigdb)

    vs.write_varstore(output_path, varlist)
    return removed


class DpkgArchDict(UserDict):
    data = {}
    dpkg_arch_out = subprocess.check_output(
        ['dpkg-architecture'],
    ).decode().rstrip().splitlines()
    for var in dpkg_arch_out:
        key, val = var.split('=')
        data[key] = val

    def __init__(self):
        pass


class SecvarUpdateTest(unittest.TestCase):
    DpkgArch = DpkgArchDict()

    def setUp(self):
        self.startTime = time.time()
        self.root = Path(os.environ.get('DEB_EDK2_ROOT', '/'))

    def tearDown(self):
        t = time.time() - self.startTime
        sys.stdout.write("%s runtime: %.3fs\n" % (self.id(), t))

    def boot_and_shutdown(self, cmd):
        """Boot QEMU, wait for shell or no-boot prompt, then shut down."""
        logger.debug(f"Running: {' '.join(cmd)}")
        child = pexpect.spawn(' '.join(cmd), encoding='UTF-8')
        child.logfile = sys.stdout
        try:
            i = child.expect(
                [
                    'Press .* or any other key to continue',
                    'BdsDxe: No bootable option or device was found',
                    pexpect.TIMEOUT,
                ],
                timeout=TEST_TIMEOUT,
            )
            if i == 0:
                child.sendline('\x1b')
                child.expect(['Shell> '], timeout=TEST_TIMEOUT)
                child.sendline('reset -s\r')
            elif i == 1:
                child.close()
        except pexpect.EOF:
            pass

    def make_2011_only_vars(self, template_path):
        """
        Create a temp VARS file with 2023 certs stripped.
        Caller must os.unlink() the returned path.
        """
        f = tempfile.NamedTemporaryFile(suffix='.fd', delete=False)
        f.close()

        self.assertTrue(
            has_cert_cn(str(template_path), 'KEK', CERT_CNS['kek_2011']),
            "2011 KEK CA missing from VARS template")

        removed = strip_2023_certs(str(template_path), f.name)
        self.assertGreater(removed, 0,
            "No 2023 certs to strip - template may be outdated")
        self.assertFalse(
            has_cert_cn(f.name, 'KEK', CERT_CNS['kek_2023']),
            "2023 KEK CA still present after stripping")

        return f.name

    def assert_all_certs_present(self, vars_path):
        """Assert all expected 2011 and 2023 certs are in KEK and db."""
        for var_name, cert_key in EXPECTED_CERTS:
            self.assertTrue(
                has_cert_cn(vars_path, var_name, CERT_CNS[cert_key]),
                f"{cert_key} missing from {var_name}")

    def boot_ovmf(self, vars_path):
        """Boot OVMF Q35 with the given VARS file. Returns the QemuCommand."""
        code = self.root / 'usr/share/OVMF/OVMF_CODE_4M.ms.fd'
        q = Qemu.QemuCommand(
            QemuEfiMachine.OVMF_Q35,
            code_path=code,
            vars_template_path=Path(vars_path),
            flash_size=QemuEfiFlashSize.SIZE_4MB,
        )
        self.boot_and_shutdown(q.command)
        return q

    @unittest.skipUnless(
        DpkgArch['DEB_BUILD_ARCH_BITS'] == '64', "Requires 64-bit QEMU",
    )
    @unittest.skipUnless(
        DpkgArch['DEB_BUILD_ARCH'] == 'amd64', "amd64-only",
    )
    def test_ovmf_secvar_update_appends_2023_certs(self):
        """Boot OVMF with 2011-only VARS, verify 2023 CAs are appended."""
        ms_vars = self.root / 'usr/share/OVMF/OVMF_VARS_4M.ms.fd'
        vars_path = self.make_2011_only_vars(ms_vars)
        try:
            q = self.boot_ovmf(vars_path)
            self.assert_all_certs_present(q.pflash.varfile.name)
        finally:
            os.unlink(vars_path)

    @unittest.skipUnless(
        DpkgArch['DEB_BUILD_ARCH_BITS'] == '64', "Requires 64-bit QEMU",
    )
    @unittest.skipUnless(
        DpkgArch['DEB_BUILD_ARCH'] == 'amd64', "amd64-only",
    )
    def test_ovmf_secvar_update_version_guard(self):
        """After the driver has run, stripping 2023 certs and rebooting
        should NOT re-add them - the version variable prevents re-running."""
        ms_vars = self.root / 'usr/share/OVMF/OVMF_VARS_4M.ms.fd'
        vars_path = self.make_2011_only_vars(ms_vars)
        try:
            # First boot - driver runs, appends 2023 certs
            q1 = self.boot_ovmf(vars_path)
            self.assert_all_certs_present(q1.pflash.varfile.name)

            # Strip 2023 certs again (simulating user removal), but keep
            # the version variable that the driver wrote
            stripped_path = q1.pflash.varfile.name + '.stripped'
            removed = strip_2023_certs(q1.pflash.varfile.name, stripped_path)
            self.assertGreater(removed, 0)

            # Second boot - driver should skip due to version guard
            q2 = self.boot_ovmf(stripped_path)

            # 2023 certs should NOT be back
            for cn in ALL_2023_CNS:
                self.assertFalse(
                    has_cert_cn(q2.pflash.varfile.name, 'KEK', cn) or
                    has_cert_cn(q2.pflash.varfile.name, 'db', cn),
                    f"Driver re-ran despite version guard: {cn} was re-added")

            os.unlink(stripped_path)
        finally:
            os.unlink(vars_path)

    @unittest.skipUnless(
        DpkgArch['DEB_BUILD_ARCH_BITS'] == '64', "Requires 64-bit QEMU",
    )
    @unittest.skipUnless(
        DpkgArch['DEB_BUILD_ARCH'] == 'arm64', "arm64-only",
    )
    def test_aavmf_secvar_update_appends_2023_certs(self):
        """Boot AAVMF with 2011-only VARS, verify 2023 CAs are appended."""
        code = self.root / 'usr/share/AAVMF/AAVMF_CODE.ms.fd'
        ms_vars = self.root / 'usr/share/AAVMF/AAVMF_VARS.ms.fd'

        vars_path = self.make_2011_only_vars(ms_vars)
        try:
            q = Qemu.QemuCommand(
                QemuEfiMachine.AAVMF,
                code_path=code,
                vars_template_path=Path(vars_path),
            )
            self.boot_and_shutdown(q.command)
            self.assert_all_certs_present(q.pflash.varfile.name)
        finally:
            os.unlink(vars_path)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logging.basicConfig(level=logging.DEBUG)
    unittest.main(verbosity=2)
